import platform
from pprint import pprint
import sys
import os
from pathlib import Path
import re
import shutil
import textwrap
from typing import Any, Dict, List, Optional
import tempfile
from subprocess import CalledProcessError, run as sp_run
from glob import glob

from . import config
from . import cmake
from .datastructures import BuildPaths, PackageInfo

import flit_core.common  # type: ignore
import flit_core.config  # type: ignore
from distlib.version import NormalizedVersion  # type: ignore

_CMAKE_MINIMUM_REQUIRED = NormalizedVersion('3.15')


class _BuildBackend(object):

    # --- Constructor ---------------------------------------------------------

    def __init__(self) -> None:
        self.verbose: bool = False

    # --- Methods required by PEP 517 -----------------------------------------

    def get_requires_for_build_wheel(self, config_settings=None):
        """https://www.python.org/dev/peps/pep-0517/#get-requires-for-build-wheel"""
        self.parse_config_settings(config_settings)
        pyproject = Path('pyproject.toml').resolve()
        cfg = self.read_config(pyproject, config_settings, self.verbose)
        deps = []
        # Check if we need CMake
        if cfg.cmake:
            self.check_cmake_program(cfg, deps)
        if cfg.stubgen:
            self.check_stubgen_program(cfg, deps)
        if self.verbose:
            print('Dependencies for build:', deps)
        return deps

    def get_requires_for_build_editable(self, config_settings=None):
        """https://www.python.org/dev/peps/pep-0660/#get-requires-for-build-editable"""
        return self.get_requires_for_build_wheel(config_settings)

    def get_requires_for_build_sdist(self, config_settings=None):
        """https://www.python.org/dev/peps/pep-0517/#get-requires-for-build-sdist"""
        return []

    def build_wheel(self,
                    wheel_directory,
                    config_settings=None,
                    metadata_directory=None):
        """https://www.python.org/dev/peps/pep-0517/#build-wheel"""
        assert metadata_directory is None

        # Parse options
        self.parse_config_settings(config_settings)

        # Build wheel
        with tempfile.TemporaryDirectory() as tmp_build_dir:
            whl_name = self.build_wheel_in_dir(wheel_directory, tmp_build_dir,
                                               config_settings)
        return whl_name

    def build_editable(self,
                       wheel_directory,
                       config_settings=None,
                       metadata_directory=None):
        """https://www.python.org/dev/peps/pep-0660/#build-editable"""
        assert metadata_directory is None

        # Parse options
        self.parse_config_settings(config_settings)

        # Build wheel
        with tempfile.TemporaryDirectory() as tmp_build_dir:
            whl_name = self.build_wheel_in_dir(wheel_directory,
                                               tmp_build_dir,
                                               config_settings,
                                               editable=True)
        return whl_name

    def build_sdist(self, sdist_directory, config_settings=None):
        """https://www.python.org/dev/peps/pep-0517/#build-sdist"""
        sdist_directory = Path(sdist_directory)
        src_dir = Path().resolve()

        # Parse options
        self.parse_config_settings(config_settings)

        # Load metadata
        from flit_core.common import Module, make_metadata
        pyproject = src_dir / 'pyproject.toml'
        cfg = self.read_config(pyproject, config_settings, self.verbose)
        import_name = cfg.module['name']
        pkg = Module(import_name, src_dir / cfg.module['directory'])
        metadata = make_metadata(pkg, cfg)
        metadata.version = self.normalize_version(metadata.version)

        # Export dist
        from flit_core.sdist import SdistBuilder  # type: ignore
        rel_pyproject = os.path.relpath(pyproject, src_dir)
        extra_files = [str(rel_pyproject)] + cfg.referenced_files
        sdist_cfg = cfg.sdist[self.get_os_name()]
        sdist_builder = SdistBuilder(
            pkg,
            metadata=metadata,
            cfgdir=src_dir,
            reqs_by_extra=None,
            entrypoints=cfg.entrypoints,
            extra_files=extra_files,
            data_directory=None,
            include_patterns=sdist_cfg.get('include_patterns', []),
            exclude_patterns=sdist_cfg.get('exclude_patterns', []),
        )
        sdist_tar = sdist_builder.build(Path(sdist_directory))
        return os.path.relpath(sdist_tar, sdist_directory)

    # --- Parsing config options and metadata ---------------------------------

    @staticmethod
    def is_verbose_enabled(config_settings: Optional[dict]):
        if 'PY_BUILD_CMAKE_VERBOSE' in os.environ:
            return True
        if config_settings is None:
            return False
        if config_settings.keys() & {'verbose', '--verbose', 'V', '-V'}:
            return True
        return False

    def parse_config_settings(self, config_settings: Optional[Dict]):
        self.verbose = self.is_verbose_enabled(config_settings)

    @staticmethod
    def read_config(pyproject_path: Path, config_settings: Optional[Dict],
                    verbose: bool) -> config.Config:
        config_settings = config_settings or {}
        try:
            if verbose:
                print("Configuration settings:")
                pprint(config_settings)
            listify = lambda x: x if isinstance(x, list) else [x]
            keys = ['--local', '--cross']
            overrides = {
                key: listify(config_settings.get(key) or [])
                for key in keys
            }
            if verbose:
                print("Configuration settings for local and "
                      "cross-compilation overrides:")
                pprint(overrides)
            cfg = config.read_config(pyproject_path, overrides)
        except flit_core.config.ConfigError as e:
            e.args = ("\n"
                      "\n"
                      "\t❌ Error in user configuration:\n"
                      "\n"
                      f"\t\t{e}\n"
                      "\n", )
            raise
        except AssertionError as e:
            e.args = (
                "\n"
                "\n"
                "\t❌ Internal error while processing the configuration\n"
                "\t   Please notify the developers: https://github.com/tttapa/py-build-cmake/issues\n"
                "\n"
                f"\t\t{e}\n"
                "\n", )
            raise
        if verbose:
            _BuildBackend.print_config_verbose(cfg)
        return cfg

    # --- Building wheels -----------------------------------------------------

    def build_wheel_in_dir(self,
                           wheel_directory_,
                           tmp_build_dir,
                           config_settings,
                           editable=False):
        """This is the main function that contains all steps necessary to build
        a complete wheel package, including the CMake builds etc."""
        # Set up all paths
        paths = BuildPaths(
            source_dir=Path().resolve(),
            wheel_dir=Path(wheel_directory_),
            temp_dir=Path(tmp_build_dir),
            staging_dir=Path(tmp_build_dir) / 'staging',
        )

        # Load metadata from the pyproject.toml file
        cfg, pkg, metadata = _BuildBackend.read_all_metadata(
            paths.source_dir, config_settings, self.verbose)

        pkg_info = PackageInfo(
            version=metadata.version,
            package_name=cfg.package_name,
            module_name=cfg.module['name'],
        )

        # Copy the module's Python source files to the temporary folder
        if not editable:
            self.copy_pkg_source_to(paths.staging_dir, pkg)
        else:
            self.write_editable_wrapper(paths.staging_dir, pkg)

        # Create dist-info folder
        distinfo_dir = f'{pkg_info.package_name}-{pkg_info.version}.dist-info'
        distinfo_dir = paths.staging_dir / distinfo_dir
        os.makedirs(distinfo_dir, exist_ok=True)

        # Write metadata
        metadata_path = distinfo_dir / 'METADATA'
        with open(metadata_path, 'w', encoding='utf-8') as f:
            metadata.write_metadata_file(f)
        # Write or copy license
        self.write_license_files(cfg.license, paths.source_dir, distinfo_dir)
        # Write entrypoints/scripts
        self.write_entrypoints(distinfo_dir, cfg)

        # Generate .pyi stubs (for the Python files only)
        if cfg.stubgen is not None and not editable:
            self.generate_stubs(paths, pkg, cfg.stubgen)

        # Configure, build and install the CMake project
        if cfg.cmake:
            self.do_native_cross_cmake_build(paths, cfg, pkg_info)

        # Create wheel
        whl_name = self.create_wheel(paths, cfg, pkg_info)
        return whl_name

    @staticmethod
    def read_all_metadata(src_dir, config_settings, verbose):
        from flit_core.common import Module, make_metadata
        cfg = _BuildBackend.read_config(src_dir / 'pyproject.toml',
                                        config_settings, verbose)
        pkg = Module(cfg.module['name'], src_dir / cfg.module['directory'])
        metadata = make_metadata(pkg, cfg)
        metadata.version = _BuildBackend.normalize_version(metadata.version)
        return cfg, pkg, metadata

    def create_wheel(self, paths, cfg, package_info):
        """Create a wheel package from the build directory."""
        from distlib.wheel import Wheel  # type: ignore
        whl = Wheel()
        whl.name = package_info.package_name
        whl.version = package_info.version
        pure = not cfg.cmake
        libdir = 'purelib' if pure else 'platlib'
        staging_dir = paths.staging_dir
        whl_paths = {'prefix': str(staging_dir), libdir: str(staging_dir)}
        whl.dirname = paths.wheel_dir
        if pure:
            tags = {'pyver': ['py3']}
        elif cfg.cross:
            tags = self.get_cross_tags(cfg.cross)
        else:
            tags = self.get_native_tags()
        wheel_path = whl.build(whl_paths, tags=tags, wheel_version=(1, 0))
        whl_name = os.path.relpath(wheel_path, paths.wheel_dir)
        return whl_name

    @staticmethod
    def get_cmake_configs(cfg):
        native_cmake_cfg = cfg.cmake[_BuildBackend.get_os_name()]
        cmake_cfg = cfg.cmake['cross'] if cfg.cross else native_cmake_cfg
        return cmake_cfg, native_cmake_cfg

    def do_native_cross_cmake_build(self, paths: BuildPaths, cfg,
                                    package_info):
        """If not cross-compiling, just do a regular CMake build+install.
        When cross-compiling, do a cross-build+install (using the provided 
        CMake toolchain file).
        If cfg.cross['copy_from_native_build'] is set, before cross-compiling, 
        first a normal build+install is performed to a separate directory, then
        the cross-build+install is performed, and finally the installed files
        from the native build that match the patterns in
        cfg.cross['copy_from_native_build'] are copied to the installation
        directory of the cross-build for packaging."""
        cmake_cfg, native_cmake_cfg = self.get_cmake_configs(cfg)
        # When cross-compiling, optionally do a native build first
        native_install_dir = None
        if self.needs_cross_native_build(cfg):
            native_install_dir = paths.temp_dir / 'native-install'
            self.run_cmake(paths.source_dir, native_install_dir,
                           native_cmake_cfg, None, package_info,
                           native_install_dir)
        # Then do the actual build
        self.run_cmake(paths.source_dir, paths.staging_dir, cmake_cfg,
                       cfg.cross, package_info, native_install_dir)
        # Finally, move the files from the native build to the staging area
        if native_install_dir:
            self.copy_native_install(paths.staging_dir, native_install_dir,
                                     cfg.cross['copy_from_native_build'])

    def copy_native_install(self, staging_dir, native_install_dir,
                            native_install_patterns):
        """Copy the files that match the patterns from the native installation
        directory to the wheel staging directory."""
        for pattern in native_install_patterns:
            matches = sorted(glob(str(native_install_dir / pattern)))
            for path in matches:
                path = Path(path)
                rel = path.relative_to(native_install_dir)
                path.parent.mkdir(parents=True, exist_ok=True)
                print('-- Moving:', path, '->', staging_dir / rel.parent)
                shutil.move(path, staging_dir / rel.parent)
                # TODO: what if the folder already exists?
            if not matches:
                raise RuntimeError(
                    "Native build installed no files that matched the "
                    "pattern '" + pattern + "'")
        shutil.rmtree(native_install_dir)

    # --- Files installation --------------------------------------------------

    def copy_pkg_source_to(self, staging_dir: Path,
                           pkg: flit_core.common.Module):
        """Copy the files of a Python package to the build directory."""
        for mod_file in pkg.iter_files():
            rel_path = os.path.relpath(mod_file, pkg.path.parent)
            dst = staging_dir / rel_path
            os.makedirs(dst.parent, exist_ok=True)
            shutil.copy2(mod_file, dst, follow_symlinks=False)

    def write_license_files(self, license, srcdir: Path, distinfo_dir: Path):
        """Write the LICENSE file from pyproject.toml to the distinfo
        directory."""
        if 'text' in license:
            with (distinfo_dir / 'LICENSE').open('w', encoding='utf-8') as f:
                f.write(license['text'])
        elif 'file' in license:
            shutil.copy2(srcdir / license['file'], distinfo_dir)

    def write_entrypoints(self, distinfo: Path, cfg: config.Config):
        from flit_core.common import write_entry_points
        with (distinfo / 'entry_points.txt').open('w', encoding='utf-8') as f:
            write_entry_points(cfg.entrypoints, f)

    # --- Editable installs ---------------------------------------------------

    def write_editable_wrapper(self, staging_dir: Path,
                               pkg: flit_core.common.Module):
        """Write a fake __init__.py file that points to the development 
        folder."""
        tmp_pkg: Path = staging_dir / pkg.name
        pkgpath = Path(pkg.path)
        initpath = pkgpath / '__init__.py'
        os.makedirs(tmp_pkg, exist_ok=True)
        special_dunders = [
            '__builtins__', '__cached__', '__file__', '__loader__', '__name__',
            '__package__', '__path__', '__spec__'
        ]
        content = f"""\
            # First extend the search path with the development folder
            __spec__.submodule_search_locations.insert(0, {str(pkgpath)!a})
            # Now manually import the development __init__.py
            from importlib import util as _util
            _spec = _util.spec_from_file_location("{pkg.name}",
                                                  {str(initpath)!a})
            _mod = _util.module_from_spec(_spec)
            _spec.loader.exec_module(_mod)
            # After importing, add its symbols to our global scope
            _vars = _mod.__dict__.copy()
            for _k in ['{"','".join(special_dunders)}']: _vars.pop(_k)
            globals().update(_vars)
            # Clean up
            del _k, _spec, _mod, _vars, _util
            """
        (tmp_pkg / '__init__.py').write_text(textwrap.dedent(content),
                                             encoding='utf-8')
        # Add the py.typed file if it exists, so mypy picks up the stubs for
        # the C++ extensions
        py_typed: Path = pkg.path / 'py.typed'
        if py_typed.exists():
            shutil.copy2(py_typed, tmp_pkg)
        # Write a path file so IDEs find the correct files as well
        (staging_dir / f'{pkg.name}.pth').write_text(str(pkg.path.parent))

    # --- Invoking CMake builds -----------------------------------------------

    def run_cmake(self, pkgdir, install_dir, cmake_cfg, cross_cfg,
                  package_info, native_install_dir):
        """Configure, build and install using CMake."""

        cmaker = self.get_cmaker(pkgdir,
                                 install_dir,
                                 cmake_cfg,
                                 cross_cfg,
                                 native_install_dir,
                                 package_info,
                                 verbose=self.verbose)

        cmaker.configure()
        cmaker.build()
        cmaker.install()

    @staticmethod
    def get_cmaker(pkg_dir: Path, install_dir: Optional[Path], cmake_cfg: dict,
                   cross_cfg: Optional[dict],
                   native_install_dir: Optional[Path],
                   package_info: PackageInfo, **kwargs):
        toolchain_file = None
        if cross_cfg:
            toolchain_file = (pkg_dir / cross_cfg['toolchain_file']).resolve()

        # Add some CMake configure options
        options = cmake_cfg.get('options', {})
        if native_install_dir:
            options['PY_BUILD_CMAKE_NATIVE_INSTALL_DIR:PATH'] = \
                str(native_install_dir)
        btype = cmake_cfg.get('build_type')
        if btype:  # -D CMAKE_BUILD_TYPE={type}
            options['CMAKE_BUILD_TYPE:STRING'] = btype

        # Build folder for each platform
        build_cfg_name = _BuildBackend.get_build_config_name(cross_cfg)

        # CMake options
        return cmake.CMaker(
            cmake_settings=cmake.CMakeSettings(
                working_dir=Path(pkg_dir),
                source_path=Path(cmake_cfg["source_path"]),
                build_path=Path(cmake_cfg['build_path']) / build_cfg_name,
                os=_BuildBackend.get_os_name(),
                command=Path("cmake"),
            ),
            conf_settings=cmake.CMakeConfigureSettings(
                environment=cmake_cfg.get("env", {}),
                toolchain_file=toolchain_file,
                build_type=cmake_cfg.get('build_type'),
                options=options,
                args=cmake_cfg.get('args', []),
                preset=cmake_cfg.get('preset'),
                generator=cmake_cfg.get('generator'),
            ),
            build_settings=cmake.CMakeBuildSettings(
                args=cmake_cfg['build_args'],
                tool_args=cmake_cfg['build_tool_args'],
                presets=cmake_cfg.get('build_presets', []),
                configs=cmake_cfg.get('config', []),
            ),
            install_settings=cmake.CMakeInstallSettings(
                args=cmake_cfg['install_args'],
                presets=cmake_cfg.get('install_presets', []),
                configs=cmake_cfg.get('config', []),
                components=cmake_cfg.get('install_components', []),
                prefix=install_dir,
            ),
            package_info=package_info,
            **kwargs,
        )

    # --- Generate stubs ------------------------------------------------------

    def generate_stubs(self, paths: BuildPaths, pkg, cfg: Dict[str, Any]):
        """Generate stubs (.pyi) using mypy stubgen."""
        args = ['stubgen'] + cfg.get('args', [])
        cfg.setdefault('packages', [pkg.name] if pkg.is_package else [])
        for p in cfg['packages']:
            args += ['-p', p]
        cfg.setdefault('modules', [pkg.name] if not pkg.is_package else [])
        for m in cfg['modules']:
            args += ['-m', m]
        args += cfg.get('files', [])
        # Add output folder argument if not already specified in cfg['args']
        if 'args' not in cfg or not ({'-o', '--output'} & set(cfg['args'])):
            args += ['-o', str(paths.staging_dir)]
        env = os.environ.copy()
        env.setdefault('MYPY_CACHE_DIR', str(paths.temp_dir))
        # Call mypy stubgen in a subprocess
        self.run(args, cwd=pkg.path.parent, check=True, env=env)

    # --- Misc helper functions -----------------------------------------------

    @staticmethod
    def get_os_name():
        return {
            "Linux": "linux",
            "Windows": "windows",
            "Darwin": "mac",  # TODO: untested
        }[platform.system()]

    @staticmethod
    def print_config_verbose(cfg):
        print("\npy-build-cmake options")
        print("======================")
        print("module:")
        pprint(cfg.module)
        print("sdist:")
        pprint(cfg.sdist)
        print("cmake:")
        pprint(cfg.cmake)
        print("stubgen:")
        pprint(cfg.stubgen)
        print("cross:")
        pprint(cfg.cross)
        print("======================\n")

    @staticmethod
    def normalize_version(version):
        from distlib.version import NormalizedVersion
        norm_version = str(NormalizedVersion(version))
        return norm_version

    def run(self, *args, **kwargs):
        """Wrapper around subprocess.run that optionally prints the command."""
        if self.verbose:
            pprint([*args])
            pprint(kwargs)
        return sp_run(*args, **kwargs)

    @staticmethod
    def get_native_tags():
        """Get the PEP 425 tags for the current platform."""
        from .tags import get_python_tag, get_abi_tag, get_platform_tag
        return {
            'pyver': [get_python_tag()],
            'abi': [get_abi_tag()],
            'arch': [get_platform_tag()],
        }

    @staticmethod
    def get_cross_tags(crosscfg):
        """Get the PEP 425 tags to use when cross-compiling."""
        tags = _BuildBackend.get_native_tags()
        if 'implementation' in crosscfg and 'version' in crosscfg:
            tags['pyver'] = [crosscfg['implementation'] + crosscfg['version']]
        if 'abi' in crosscfg:
            tags['abi'] = [crosscfg['abi']]
        if 'arch' in crosscfg:
            tags['arch'] = [crosscfg['arch']]
        return tags

    @staticmethod
    def get_build_config_name(cross_cfg):
        """Get a string representing the Python version, ABI and architecture,
        used to name the build folder so builds for different versions don't
        interfere."""
        if cross_cfg:
            tags = _BuildBackend.get_cross_tags(cross_cfg)
        else:
            tags = _BuildBackend.get_native_tags()
        return '-'.join(map(lambda x: x[0], tags.values()))

    @staticmethod
    def needs_cross_native_build(cfg):
        return cfg.cross and 'copy_from_native_build' in cfg.cross

    @staticmethod
    def iter_files(stagedir):
        """Iterate over the files contained in the given folder.

        Yields absolute paths - caller may want to make them relative.
        Excludes any __pycache__ and *.pyc files."""

        # https://github.com/pypa/flit/blob/a4524758604107bde8c77b5816612edb76a604aa/flit_core/flit_core/common.py#L73

        def _include(path):
            name = os.path.basename(path)
            return name != '__pycache__' and not name.endswith('.pyc')

        # Ensure we sort all files and directories so the order is stable
        for dirpath, dirs, files in os.walk(str(stagedir)):
            for file in sorted(files):
                full_path = os.path.join(dirpath, file)
                if _include(full_path):
                    yield full_path

            dirs[:] = [d for d in sorted(dirs) if _include(d)]

    # --- Helper functions for finding programs like CMake --------------------

    def check_cmake_program(self, cfg: config.Config, deps: List[str]):
        assert cfg.cmake
        # Do we need to perform a native build?
        native = not cfg.cross or self.needs_cross_native_build(cfg)
        native_cfg = cfg.cmake[self.get_os_name()] if native else {}
        # Do we need to perform a cross build?
        cross = cfg.cross
        cross_cfg = cfg.cmake.get('cross', {})
        # Find the strictest version requirement
        min_cmake_ver = max(
            _CMAKE_MINIMUM_REQUIRED,
            NormalizedVersion(native_cfg.get('minimum_version', '0.0')),
            NormalizedVersion(cross_cfg.get('minimum_version', '0.0')),
        )
        # If CMake in PATH doesn't work or is too old, add it as a build
        # requirement
        if not self.check_program_version('cmake', min_cmake_ver, "CMake"):
            deps.append("cmake>=" + str(min_cmake_ver))

        # Check if we need Ninja
        cfgs = []
        if native: cfgs.append(native_cfg)
        if cross: cfgs.append(cross_cfg)
        # Do any of the configs require Ninja as a generator?
        needs_ninja = lambda c: 'ninja' in c.get('generator', '').lower()
        need_ninja = any(map(needs_ninja, cfgs))
        if need_ninja:
            # If so, check if a working version exists in the PATH, otherwise,
            # add it as a build requirement
            if not self.check_program_version('ninja', None, "Ninja"):
                deps.append("ninja")

    def check_stubgen_program(self, cfg: config.Config, deps: List[str]):
        if not self.check_program_version('stubgen', None, None, False):
            deps.append("mypy")

    def check_program_version(
        self,
        program: str,
        minimum_version: Optional[NormalizedVersion],
        name: Optional[str],
        check_version: bool = True,
    ):
        """Check if there's a new enough version of the given command available
        in PATH."""
        name = name or program
        try:
            # Try running the command
            cmd = [program, '--version'] if check_version else [program, '-h']
            res = self.run(cmd, check=True, capture_output=True)
            # Try finding the version
            if check_version:
                m = re.search(r'\d+(\.\d+){1,}', res.stdout.decode('utf-8'))
                if not m:
                    raise RuntimeError(f"Unexpected {name} version output")
                program_version = NormalizedVersion(m.group(0))
                if self.verbose: print("Found", name, program_version)
                # Check if the version is new enough
                if minimum_version is not None:
                    if program_version < minimum_version:
                        raise RuntimeError(f"{name} too old")
        except CalledProcessError as e:
            if self.verbose:
                print(f'{type(e).__module__}.{type(e).__name__}', e, sep=': ')
                sys.stdout.buffer.write(e.stdout)
                sys.stdout.buffer.write(e.stderr)
            return False
        except Exception as e:
            # If any of that failed, return False
            if self.verbose:
                print(f'{type(e).__module__}.{type(e).__name__}', e, sep=': ')
            return False
        return True


_BACKEND = _BuildBackend()
get_requires_for_build_wheel = _BACKEND.get_requires_for_build_wheel
get_requires_for_build_sdist = _BACKEND.get_requires_for_build_sdist
get_requires_for_build_editable = _BACKEND.get_requires_for_build_editable
build_wheel = _BACKEND.build_wheel
build_sdist = _BACKEND.build_sdist
build_editable = _BACKEND.build_editable



import html
import itertools
import shutil
import textwrap

from ..config_options import ConfigOption, PathConfigOption, RequiredValue, pth2str


def _print_wrapped(text, indent, width=None):
    """Print the given string with the given indentation, wrapping it to the
    desired width, preserving line endings. If `width` is None, use the width 
    of the terminal, or 80 columns as a fallback."""
    if width is None:
        width = shutil.get_terminal_size((80, 24)).columns
    wrapper = textwrap.TextWrapper(width=width,
                                   initial_indent=indent,
                                   subsequent_indent=indent)
    wrapped = [wrapper.wrap(i) for i in text.split('\n')]
    for line in itertools.chain.from_iterable(wrapped):
        print(line)


def get_default_str(opt: ConfigOption):
    """Get a string representation of the default value for the given option."""
    return opt.default.get_name()


def help_print_md(pbc_opts: ConfigOption):
    """
    Prints the top-level options in `pbc_opts` as MarkDown tables.
    """
    print("# py-build-cmake configuration options\n")
    print("These options go in the `[tool.py-build-cmake]` section of "
          "the `pyproject.toml` configuration file.\n")
    for k, v in pbc_opts.sub.items():
        print('##', k)
        print(_get_full_description(v), '\n')
        print('| Option | Description | Type | Default |')
        print('|--------|-------------|------|---------|')
        for kk, vv in v.sub.items() or {}:
            print('|', f'`{kk}`', '|', _get_full_description(vv), '|',
                  vv.get_typename() or '', '|', f'`{get_default_str(vv)}`',
                  '|')
        print()
    print("# Local overrides\n")
    print("Additionally, two extra configuration files can be placed in "
          "the same directory as `pyproject.toml` to override some "
          "options for your specific use case:\n\n"
          "- `py-build-cmake.local.toml`: the options in this file "
          "override the values in the `tool.py-build-cmake` section of "
          "`pyproject.toml`.<br/>This is useful if you need specific "
          "arguments or CMake options to compile the package on your "
          "system.\n"
          "- `py-build-cmake.cross.toml`: the options in this file "
          "override the values in the `tool.py-build-cmake.cross` section "
          "of `pyproject.toml`.<br/>Useful for cross-compiling the "
          "package without having to edit the main configuration file.\n")
    print("# Command line overrides\n")
    print("Instead of using the `py-build-cmake.local.toml` and "
          "`py-build-cmake.cross.toml` files, you can also include "
          "additional config files using command line options:\n\n"
          "- `--local`: specifies a toml file that overrides the "
          "`tool.py-build-cmake` section of `pyproject.toml`, "
          "similar to `py-build-cmake.local.toml`\n"
          "- `--cross`: specifies a toml file that overrides the "
          "`tool.py-build-cmake.cross` section of `pyproject.toml`, "
          "similar to `py-build-cmake.cross.toml`\n\n"
          "These command line overrides are applied after the "
          "`py-build-cmake.local.toml` and `py-build-cmake.cross.toml` "
          "files in the project folder (if any).\n\n"
          "When using PyPA build, these flags can be specified using "
          "the `-C` or `--config-setting` flag: \n"
          "```sh\n"
          "python -m build . -C--cross=/path/to/my-cross-config.toml\n"
          "```\n"
          "The same flag may appear multiple times, for example: \n"
          "```sh\n"
          "python -m build . -C--local=conf-A.toml -C--local=conf-B.toml\n"
          "```\n"
          "For PyPA pip, you can use the `--config-settings` flag "
          "instead.")


def _get_full_description(vv: ConfigOption):
    descr = _md_escape(vv.description)
    if isinstance(vv, PathConfigOption):
        descr += '<br/>' + _describe_path_option(vv).capitalize() + '.'
    if vv.inherit_from:
        descr += '<br/>Inherits from: `/' + pth2str(vv.inherit_from) + '`'
    if vv.example:
        descr += '<br/>For example: `' + vv.example + '`'
    return descr


def _md_escape(descr):
    descr = html.escape(descr)
    descr = descr.replace('\n', '<br/>')
    descr = descr.replace('*', '\\*')
    descr = descr.replace('_', '\\_')
    return descr


def recursive_help_print(opt: ConfigOption, level=0):
    """Recursively prints the help messages for the options in `opt`."""
    for k, v in opt.sub.items():
        if k == 'project':
            continue
        indent = 4 * level * ' '
        header = '\n' + k
        if v.sub:
            header += ':'
            print(textwrap.indent(header, indent))
            if v.description:
                _print_wrapped(v.description, indent + '  ')
            recursive_help_print(v, level + 1)
        else:
            headerfields = []
            typename = v.get_typename()
            if typename is not None:
                headerfields += [typename]
            if isinstance(v, PathConfigOption):
                headerfields += [_describe_path_option(v)]
            is_required = isinstance(v.default, RequiredValue)
            if is_required:
                headerfields += ['required']
            if v.inherit_from:
                headerfields += ['inherits from /' + pth2str(v.inherit_from)]
            if headerfields:
                header += ' (' + ', '.join(headerfields) + ')'
            print(textwrap.indent(header + ':', indent))
            _print_wrapped(v.description, indent + '  ')
            if v.example:
                _print_wrapped('For example: ' + v.example, indent + '  ')
            default = v.default.get_name()
            if default is not None and not is_required:
                print(textwrap.indent('Default: ' + default, indent + '  '))


def _describe_path_option(v: PathConfigOption):
    t = 'absolute or relative' if v.allow_abs else 'relative'
    if v.base_path is not None:
        t += ' to ' + v.base_path.description
    return t


def help_print(pbc_opts: ConfigOption):
    print("List of py-build-cmake pyproject.toml options:")
    recursive_help_print(pbc_opts)
    print()



from . import __version__
from pathlib import Path
import click


def cmake_command(directory, build_path, verbose, dry, native, cross, local):

    def get_cmaker():
        from .build import _BuildBackend as backend
        from .datastructures import PackageInfo
        source_dir = Path(directory or '.').resolve()
        config_settings = {
            "--cross": cross,
            "--local": local,
        }
        # Read configuration and package metadata
        cfg, pkg, metadata = backend.read_all_metadata(source_dir,
                                                       config_settings,
                                                       verbose)

        if not cfg.cmake:
            print("Not a CMake package")
            return None

        pkg_info = PackageInfo(
            version=metadata.version,
            package_name=cfg.package_name,
            module_name=cfg.module["name"],
        )
        # Select the right configuration (native build or cross build)
        cmake_cfg, native_cmake_cfg = backend.get_cmake_configs(cfg)
        cmake_cfg = native_cmake_cfg if native else cmake_cfg
        cross_cfg = None if native else cfg.cross
        # Override the build folder
        if build_path is not None:
            cmake_cfg['build_path'] = str(build_path)
        # Configure all CMake options
        return backend.get_cmaker(pkg_dir=source_dir,
                                  install_dir=None,
                                  cmake_cfg=cmake_cfg,
                                  cross_cfg=cross_cfg,
                                  native_install_dir=None,
                                  package_info=pkg_info,
                                  verbose=verbose,
                                  dry=dry)

    return get_cmaker


@click.group()
@click.option("-C",
              "--directory",
              type=click.Path(exists=True, file_okay=False, dir_okay=True),
              required=False,
              help="The directory containing pyproject.toml.")
@click.option("-B",
              "--build-path",
              type=click.Path(exists=False, file_okay=False, dir_okay=True),
              required=False,
              default=None,
              help="Override py-build-cmake's default build folder.")
@click.option("-V",
              "--verbose",
              is_flag=True,
              help="Print verbose information about the commands being "
              "executed.")
@click.option("-n",
              "--dry",
              is_flag=True,
              help="Print the commands without actually invoking CMake.")
@click.option("--native",
              is_flag=True,
              help="When the configuration requests cross-compiling, "
              "configure for a native build instead. Has no effect otherwise.")
@click.option("--local",
              type=click.Path(exists=False, file_okay=True, dir_okay=False),
              required=False,
              help="Specifies a toml file that overrides the "
              "tool.py-build-cmake section of pyproject.toml, "
              "similar to py-build-cmake.local.toml.")
@click.option("--cross",
              type=click.Path(exists=False, file_okay=True, dir_okay=False),
              required=False,
              help="Specifies a toml file that overrides the "
              "tool.py-build-cmake.cross section of pyproject.toml, "
              "similar to py-build-cmake.cross.toml.")
@click.version_option(__version__, "-v", "--version")
@click.pass_context
def cli(ctx: click.Context, **kwargs):
    ctx.obj = cmake_command(**kwargs)


@cli.command(help="Configure the CMake project.")
@click.pass_obj
@click.option("--preset", nargs=1, type=str, required=False, metavar="PRESET")
@click.argument("args", nargs=-1, required=False)
def configure(obj, preset, args):
    cmaker = obj()
    if cmaker is None:
        return
    cmaker.conf_settings.args += args or []
    if preset is not None:
        cmaker.conf_settings.preset = preset
    cmaker.configure()


@cli.command(help="Build the CMake project.")
@click.pass_obj
@click.option("--preset",
              nargs=1,
              multiple=True,
              type=str,
              required=False,
              metavar="PRESET")
@click.option("--config",
              nargs=1,
              multiple=True,
              type=str,
              required=False,
              metavar="CONFIG")
@click.argument("args", nargs=-1, required=False)
def build(obj, preset, config, args):
    cmaker = obj()
    if cmaker is None:
        return
    cmaker.build_settings.args += args or []
    if preset or config:
        cmaker.build_settings.presets = []
        cmaker.build_settings.configs = []
    if preset:
        cmaker.build_settings.presets = preset
    if config:
        cmaker.build_settings.configs = config
    cmaker.build()


@cli.command(help="Install the CMake project.")
@click.pass_obj
@click.option("--preset",
              nargs=1,
              multiple=True,
              type=str,
              required=False,
              metavar="PRESET")
@click.option("--config",
              nargs=1,
              multiple=True,
              type=str,
              required=False,
              metavar="CONFIG")
@click.option("--component",
              nargs=1,
              multiple=True,
              type=str,
              required=False,
              metavar="COMP")
@click.argument("args", nargs=-1, required=False)
def install(obj, preset, config, component, args):
    cmaker = obj()
    if cmaker is None:
        return
    cmaker.install_settings.args += args or []
    if preset or config:
        cmaker.install_settings.presets = []
        cmaker.install_settings.configs = []
    if preset:
        cmaker.install_settings.presets = preset
    if config:
        cmaker.install_settings.configs = config
    if component:
        cmaker.install_settings.components = component
    cmaker.install()


@cli.group(help="Config file operations.")
def config():
    pass


@config.command(help="Print documentation for the config file format.")
@click.option("--md", is_flag=True, help="Use the MarkDown format.")
def format(md):
    from .pyproject_options import get_options
    from .config_options import pth
    opts = get_options(Path('/'))
    help_pth = pth('pyproject.toml/tool/py-build-cmake')
    pbc_opts = opts[help_pth]
    if md:
        from .help import help_print_md as help_print
    else:
        from .help import help_print as help_print
    help_print(pbc_opts)


if __name__ == '__main__':
    cli()



import sys
import textwrap
from pathlib import Path

from ..pyproject_options import get_options
from ..config_options import pth
from . import help_print_md, help_print


def _print_usage():
    print(
        textwrap.dedent("""\
    Print the documentation for the pyproject.toml configuration options for the
    py-build-cmake build backend.

    Usage:
    """))
    name = __loader__.name.rsplit(sep='.', maxsplit=1)[0]
    print("   ", sys.executable, "-m", name, "[md]")
    print(
        textwrap.dedent("""
    Use the 'md' option to print the options as a MarkDown table instead of a
    plain-text list.
    """))


def main():
    opts = get_options(Path('/'))
    help_pth = pth('pyproject.toml/tool/py-build-cmake')
    pbc_opts = opts[help_pth]
    help_opt = {'-h', '-?', '--help', 'h', 'help', '?'}
    if len(sys.argv) == 2 and sys.argv[1] == 'md':
        help_print_md(pbc_opts)
    elif len(sys.argv) > 1 or set(map(str.lower, sys.argv[1:])) & help_opt:
        _print_usage()
    else:
        help_print(pbc_opts)


if __name__ == '__main__':
    main()


from pathlib import Path
from dataclasses import dataclass
import os
from string import Template
import sys
from typing import Dict, List, Optional
import re
from pprint import pprint
from subprocess import run as sp_run
from .datastructures import PackageInfo

@dataclass
class CMakeSettings:
    working_dir: Path
    source_path: Path
    build_path: Path
    os: str
    command: Path = Path("cmake")


@dataclass
class CMakeConfigureSettings:
    environment: dict
    toolchain_file: Optional[Path]
    build_type: Optional[str]
    options: Dict[str, str]
    args: List[str]
    preset: Optional[str]
    generator: Optional[str]


@dataclass
class CMakeBuildSettings:
    args: List[str]
    tool_args: List[str]
    presets: List[str]
    configs: List[str]


@dataclass
class CMakeInstallSettings:
    args: List[str]
    presets: List[str]
    configs: List[str]
    components: List[str]
    prefix: Optional[Path]


class CMaker:

    def __init__(self,
                 cmake_settings: CMakeSettings,
                 conf_settings: CMakeConfigureSettings,
                 build_settings: CMakeBuildSettings,
                 install_settings: CMakeInstallSettings,
                 package_info: PackageInfo,
                 verbose: bool = False,
                 dry: bool = False):
        self.cmake_settings = cmake_settings
        self.conf_settings = conf_settings
        self.build_settings = build_settings
        self.install_settings = install_settings
        self.package_info = package_info
        self.verbose = verbose
        self.dry = dry
        self.environment: Optional[dict] = None

    def run(self, *args, **kwargs):
        """Wrapper around subprocess.run that optionally prints the command."""
        if self.verbose:
            pprint([*args])
            pprint(kwargs)
        elif self.dry:
            from shlex import join
            print(join(args[0]))
        if not self.dry:
            return sp_run(*args, **kwargs)

    def prepare_environment(self):
        """Copy of the current environment with the variables defined in the
        user's configuration settings."""
        if self.environment is None:
            self.environment = os.environ.copy()
            if self.conf_settings.environment:
                for k, v in self.conf_settings.environment.items():
                    templ = Template(v)
                    self.environment[k] = templ.substitute(self.environment)
        return self.environment

    def cross_compiling(self) -> bool:
        return self.conf_settings.toolchain_file is not None

    def get_configure_options_package(self) -> List[str]:
        """Flags specific to py-build-cmake, useful in the user's CMake scripts."""
        return [
            'PY_BUILD_CMAKE_PACKAGE_VERSION:STRING=' +
            self.package_info.version,
            'PY_BUILD_CMAKE_PACKAGE_NAME:STRING=' +
            self.package_info.package_name,
            'PY_BUILD_CMAKE_MODULE_NAME:STRING=' +
            self.package_info.module_name,
        ]

    def get_configure_options_python(self) -> List[str]:
        """Flags to help CMake find the right version of Python."""
        opts = ['Python3_EXECUTABLE:FILEPATH=' + sys.executable]
        if not self.cross_compiling():
            opts += [
                'Python3_ROOT_DIR:PATH=' + sys.prefix,
                'Python3_FIND_REGISTRY=NEVER',
                'Python3_FIND_STRATEGY=LOCATION',
            ]
        return opts

    def get_configure_options_env(self, env) -> List[str]:
        """Currently sets CMAKE_OSX_ARCHITECTURES if $ENV{ARCHFLAGS} is set
        on macOS. This ensures compatibility with cibuildwheel."""
        opts = []
        if self.cmake_settings.os == "mac" and "ARCHFLAGS" in env:
            archs = re.findall(r"-arch (\S+)", env["ARCHFLAGS"])
            if archs:
                opts += ['CMAKE_OSX_ARCHITECTURES={}'.format(";".join(archs))]
        return opts

    def get_configure_options_toolchain(self) -> List[str]:
        """Sets CMAKE_TOOLCHAIN_FILE."""
        return [
            'CMAKE_TOOLCHAIN_FILE:FILEPATH=' +
            str(self.conf_settings.toolchain_file)
        ] if self.conf_settings.toolchain_file else []

    def get_configure_options_settings(self) -> List[str]:
        return [k + '=' + v for k, v in self.conf_settings.options.items()]

    def get_configure_options(self, env) -> List[str]:
        return (self.get_configure_options_package() +
                self.get_configure_options_python() +
                self.get_configure_options_env(env) +
                self.get_configure_options_toolchain() +
                self.get_configure_options_settings())

    def get_configure_command(self, env):
        options = self.get_configure_options(env)
        cmd = [str(self.cmake_settings.command)]
        cmd += ['-S', str(self.cmake_settings.source_path)]
        if self.conf_settings.preset:
            cmd += ['--preset', self.conf_settings.preset]
        else:
            cmd += ['-B', str(self.cmake_settings.build_path)]
        if self.conf_settings.generator:
            cmd += ['-G', self.conf_settings.generator]
        cmd += [f for opt in options for f in ('-D', opt)]
        cmd += self.conf_settings.args
        return cmd

    def configure(self):
        env = self.prepare_environment()
        cmd = self.get_configure_command(env)
        cwd = self.cmake_settings.working_dir
        cwd = str(cwd) if cwd is not None else None
        self.run(cmd, cwd=cwd, check=True, env=env)

    def iter_presets_configs(self, settings, func):
        done = False
        for preset in settings.presets:
            yield from func(None, preset)
            done = True
        for config in settings.configs:
            yield from func(config, None)
            done = True
        if not done:
            yield from func(None, None)

    def get_build_command(self, config, preset):
        cmd = [str(self.cmake_settings.command), '--build']
        if preset is not None:
            cmd += ['--preset', preset]
        else:
            cmd += [str(self.cmake_settings.build_path)]
        if config is not None:
            cmd += ['--config', config]
        if self.build_settings.args:
            cmd += self.build_settings.args
        if self.build_settings.tool_args:
            cmd += ['--'] + self.build_settings.tool_args
        yield cmd

    def get_build_commands(self):
        yield from self.iter_presets_configs(self.build_settings,
                                             self.get_build_command)

    def build(self):
        env = self.prepare_environment()
        cwd = self.cmake_settings.working_dir
        cwd = str(cwd) if cwd is not None else None
        for cmd in self.get_build_commands():
            self.run(cmd, cwd=cwd, check=True, env=env)

    def get_install_command(self, config, preset):
        for component in self.install_settings.components:
            cmd = [str(self.cmake_settings.command), '--install']
            if preset is not None:
                cmd += ['--preset', preset]
            else:
                cmd += [str(self.cmake_settings.build_path)]
            if config is not None:
                cmd += ['--config', config]
            if component:
                cmd += ['--component', component]
            if self.install_settings.prefix:
                cmd += ['--prefix', str(self.install_settings.prefix)]
            if self.install_settings.args:
                cmd += self.install_settings.args
            yield cmd

    def get_install_commands(self):
        yield from self.iter_presets_configs(self.install_settings,
                                             self.get_install_command)
    def install(self):
        env = self.prepare_environment()
        cwd = self.cmake_settings.working_dir
        cwd = str(cwd) if cwd is not None else None
        for cmd in self.get_install_commands():
            self.run(cmd, cwd=cwd, check=True, env=env)


from dataclasses import dataclass, field
import re
import os
import warnings
from typing import Any, Dict, List, Optional, Set
from pathlib import Path
from flit_core.config import ConfigError, read_pep621_metadata  # type: ignore
from flit_core.config import _check_glob_patterns  # type: ignore
from distlib.util import normalize_name  # type: ignore

from .config_options import ConfigNode, OverrideConfigOption
from .pyproject_options import get_options, get_cross_path, get_tool_pbc_path

try:
    import tomllib as toml_  # type: ignore
except ImportError:
    import tomli as toml_  # type: ignore


@dataclass
class Config:
    dynamic_metadata: Set[str] = field(default_factory=set)
    entrypoints: Dict[str, Dict[str, str]] = field(default_factory=dict)
    metadata: Dict[str, Any] = field(default_factory=dict)
    referenced_files: List[str] = field(default_factory=list)
    package_name: str = field(default='')
    module: Dict[str, str] = field(default_factory=dict)
    sdist: Dict[str, Dict[str, str]] = field(default_factory=dict)
    license: Dict[str, str] = field(default_factory=dict)
    cmake: Optional[Dict[str, Any]] = field(default=None)
    stubgen: Optional[Dict[str, Any]] = field(default=None)
    cross: Optional[Dict[str, Any]] = field(default=None)


def read_config(pyproject_path, flag_overrides: Dict[str,
                                                     List[str]]) -> Config:
    # Load the pyproject.toml file
    pyproject_path = Path(pyproject_path)
    pyproject_folder = pyproject_path.parent
    pyproject = toml_.loads(pyproject_path.read_text('utf-8'))
    if 'project' not in pyproject:
        raise ConfigError('Missing [project] table')

    # Load local override
    localconfig_fname = 'py-build-cmake.local.toml'
    localconfig_path = pyproject_folder / localconfig_fname
    localconfig = None
    if localconfig_path.exists():
        localconfig = toml_.loads(localconfig_path.read_text('utf-8'))
        # treat empty local override as no local override
        localconfig = localconfig or None

    # Load override for cross-compilation
    crossconfig_fname = 'py-build-cmake.cross.toml'
    crossconfig_path = pyproject_folder / crossconfig_fname
    crossconfig = None
    if crossconfig_path.exists():
        crossconfig = toml_.loads(crossconfig_path.read_text('utf-8'))
        crossconfig = crossconfig or None

    # File names mapping to the actual dict with the config
    config_files = {
        "pyproject.toml": pyproject,
        localconfig_fname: localconfig,
        crossconfig_fname: crossconfig,
    }
    # Additional options for config_options
    extra_options: List[OverrideConfigOption] = []

    def try_load_local(path: Path):
        if not path.exists():
            raise FileNotFoundError(path.absolute())
        return toml_.loads(path.read_text('utf-8'))

    extra_flag_paths = {
        '--local': get_tool_pbc_path(),
        '--cross': get_cross_path(),
    }

    for flag, targetpath in extra_flag_paths.items():
        for path in map(Path, flag_overrides[flag]):
            if not path.is_absolute():
                path = (Path(os.environ.get("PWD", ".")) / path).resolve()
            extra_options.append(
                OverrideConfigOption(
                    str(path),
                    "Command line override flag",
                    targetpath=targetpath,
                ))
            config_files[str(path)] = try_load_local(path)

    return check_config(pyproject_path, pyproject, config_files, extra_options)


def check_config(pyproject_path, pyproject, config_files, extra_options):
    # Check the package/module name and normalize it
    f = 'name'
    if f in pyproject['project']:
        normname = normalize_name(pyproject['project'][f])
        if pyproject['project'][f] != normname:
            warnings.warn(
                f"Name changed from {pyproject['project'][f]} to {normname}")
        pyproject['project'][f] = normname

    # Parse the [project] section for metadata (using flit's parser)
    flit_cfg = read_pep621_metadata(pyproject['project'], pyproject_path)

    # Create our own config data structure using flit's output
    cfg = Config()
    cfg.dynamic_metadata = flit_cfg.dynamic_metadata
    cfg.entrypoints = flit_cfg.entrypoints
    cfg.metadata = flit_cfg.metadata
    cfg.referenced_files = flit_cfg.referenced_files
    cfg.license = pyproject['project'].setdefault('license', {})
    cfg.package_name = normalize_name_wheel(cfg.metadata["name"])

    opts = get_options(pyproject_path.parent)
    for o in extra_options:
        opts.insert(o)

    tree_config = ConfigNode.from_dict(config_files)
    opts.verify_all(tree_config)
    opts.override_all(tree_config)
    opts.inherit_all(tree_config)
    opts.update_default_all(tree_config)
    dictcfg = tree_config.to_dict()
    tool_cfg = dictcfg['pyproject.toml']['tool']['py-build-cmake']

    # Store the module configuration
    s = 'module'
    if s in tool_cfg:
        # Normalize the import and wheel name of the package
        normname = normalize_name_wheel(tool_cfg[s]['name'])
        if tool_cfg[s]['name'] != normname:
            print(f"Name changed from {tool_cfg[s]['name']} to {normname}")
            # TODO: use logging instead of print
        tool_cfg[s]['name'] = normname
        cfg.module = tool_cfg[s]
    else:
        assert False, "Missing [tools.py-build-cmake.module] section"

    # Store the sdist folders (this is based on flit)
    def get_sdist_cludes(cfg):
        return {
            cl + '_patterns': _check_glob_patterns(cfg['sdist'][cl], cl)
            for cl in ('include', 'exclude')
        }

    cfg.sdist = {
        os: get_sdist_cludes(tool_cfg[os])
        for os in ("linux", "windows", "mac")
    }
    s = 'cross'
    if s in tool_cfg:
        cfg.sdist.update({
            'cross': get_sdist_cludes(tool_cfg[s]),
        })

    # Store the CMake configuration
    cfg.cmake = {
        os: tool_cfg[os]['cmake']
        for os in ("linux", "windows", "mac")
        if os in tool_cfg and 'cmake' in tool_cfg[os]
    }

    # Store stubgen configuration
    s = 'stubgen'
    if s in tool_cfg:
        cfg.stubgen = tool_cfg[s]

    # Store the cross compilation configuration
    s = 'cross'
    if s in tool_cfg:
        cfg.cross = tool_cfg[s]
        f = 'copy_from_native_build'
        if f in cfg.cross:
            cfg.cross[f] = _check_glob_patterns(cfg.cross[f], f'cross.{f}')
        cfg.cmake.update({
            'cross': cfg.cross.pop('cmake', {}),
        })

    return cfg


def normalize_name_wheel(name):
    """https://www.python.org/dev/peps/pep-0427/#escaping-and-unicode"""
    return re.sub(r"[^\w\d.]+", "_", name, re.UNICODE)



from .config_options import *


def get_tool_pbc_path():
    return pth('pyproject.toml/tool/py-build-cmake')


def get_cross_path():
    return pth('pyproject.toml/tool/py-build-cmake/cross')


def get_options(project_path: Path, *, test: bool = False):
    root = ConfigOption("root")
    pyproject = root.insert(UncheckedConfigOption("pyproject.toml"))
    project = pyproject.insert(UncheckedConfigOption('project'))
    project.insert(UncheckedConfigOption('name', default=RequiredValue()))
    name_pth = pth('pyproject.toml/project/name')
    tool = pyproject.insert(
        UncheckedConfigOption("tool",
                              default=DefaultValueValue({}),
                              create_if_inheritance_target_exists=True))
    pbc = tool.insert(
        ConfigOption("py-build-cmake",
                     default=DefaultValueValue({}),
                     create_if_inheritance_target_exists=True))

    # [tool.py-build-cmake.module]
    module = pbc.insert(
        ConfigOption(
            "module",
            "Defines the name and the directory of the module to package.",
            default=DefaultValueValue({}),
        ))
    pbc_pth = get_tool_pbc_path()
    module.insert_multiple([
        StrConfigOption('name',
                        "Import name in Python (can be different from the "
                        "name on PyPI, which is defined in the [project] "
                        "section).",
                        default=RefDefaultValue(name_pth)),
        PathConfigOption('directory',
                         "Directory containing the Python package.",
                         default=DefaultValueValue("."),
                         base_path=RelativeToProject(project_path),
                         must_exist=not test),
    ])

    # [tool.py-build-cmake.sdist]
    sdist = pbc.insert(
        ConfigOption(
            "sdist",
            "Specifies the files that should be included in the source "
            "distribution for this package.",
            default=DefaultValueValue({}),
            create_if_inheritance_target_exists=True,
        ))
    sdist_pth = pth('pyproject.toml/tool/py-build-cmake/sdist')
    sdist.insert_multiple([
        ListOfStrConfigOption('include',
                              "Files and folders to include in the source "
                              "distribution. May include the '*' wildcard "
                              "(but not '**' for recursive patterns).",
                              default=DefaultValueValue([])),
        ListOfStrConfigOption('exclude',
                              "Files and folders to exclude from the source "
                              "distribution. May include the '*' wildcard "
                              "(but not '**' for recursive patterns).",
                              default=DefaultValueValue([])),
    ])  # yapf: disable

    # [tool.py-build-cmake.cmake]
    cmake = pbc.insert(
        ConfigOption(
            "cmake",
            "Defines how to build the project to package. If omitted, "
            "py-build-cmake will produce a pure Python package.",
        ))
    cmake_pth = pth('pyproject.toml/tool/py-build-cmake/cmake')
    cmake.insert_multiple([
        StrConfigOption('minimum_version',
                        "Minimum required CMake version.",
                        "minimum_version = \"3.18\"",
                        default=NoDefaultValue()),
        StrConfigOption('build_type',
                        "Build type passed to the configuration step, as "
                        "-DCMAKE_BUILD_TYPE=<?>.",
                        "build_type = \"RelWithDebInfo\""),
        ListOfStrConfigOption('config',
                              "Configuration type passed to the build and "
                              "install steps, as --config <?>. You can "
                              "specify either a single string, or a list of "
                              "strings. If a multi-config generator is used, "
                              "all configurations in this list will be "
                              "included in the package.",
                              default=RefDefaultValue(pth('build_type'),
                                                      relative=True),
                              convert_str_to_singleton=True),
        StrConfigOption('preset',
                        "CMake preset to use for configuration. Passed as "
                        "--preset <?> during the configuration phase."),
        ListOfStrConfigOption('build_presets',
                              "CMake presets to use for building. Passed as "
                              "--preset <?> during the build phase, once for "
                              "each preset.",
                              default=RefDefaultValue(pth('preset'),
                                                      relative=True),
                              convert_str_to_singleton=True),
        ListOfStrConfigOption('install_presets',
                              "CMake presets to use for installing. Passed as "
                              "--preset <?> during the installation phase, "
                              "once for each preset.",
                              default=RefDefaultValue(pth('build_presets'),
                                                      relative=True),
                              convert_str_to_singleton=True),
        StrConfigOption('generator',
                        "CMake generator to use, passed to the "
                        "configuration step, as "
                        "-G <?>."),
        PathConfigOption('source_path',
                         "Folder containing CMakeLists.txt.",
                         default=DefaultValueValue("."),
                         expected_contents=[] if test else ["CMakeLists.txt"],
                         base_path=RelativeToProject(project_path),
                         must_exist=not test),
        PathConfigOption('build_path',
                         "CMake build and cache folder.",
                         default=DefaultValueValue('.py-build-cmake_cache'),
                         allow_abs=True,
                         base_path=RelativeToProject(project_path),
                         must_exist=False),
        DictOfStrConfigOption('options',
                              "Extra options passed to the configuration step, "
                              "as -D<option>=<value>.",
                              "options = {\"WITH_FEATURE_X\" = \"On\"}",
                              default=DefaultValueValue({})),
        ListOfStrConfigOption('args',
                              "Extra arguments passed to the configuration "
                              "step.",
                              "args = [\"--debug-find\", \"-Wdev\"]",
                              default=DefaultValueValue([])),
        ListOfStrConfigOption('build_args',
                              "Extra arguments passed to the build step.",
                              "build_args = [\"-j\"]",
                              default=DefaultValueValue([])),
        ListOfStrConfigOption('build_tool_args',
                              "Extra arguments passed to the build tool in the "
                              "build step (e.g. to Make or Ninja).",
                              "build_tool_args = [\"VERBOSE=1\"]",
                              default=DefaultValueValue([])),
        ListOfStrConfigOption('install_args',
                              "Extra arguments passed to the install step.",
                              "install_args = [\"--strip\"]",
                              default=DefaultValueValue([])),
        ListOfStrConfigOption("install_components",
                              "List of components to install, the install step "
                              "is executed once for each component, with the "
                              "option --component <?>.\n"
                              "Use an empty string to specify the default "
                              "component.",
                              default=DefaultValueValue([""])),
        DictOfStrConfigOption("env",
                              "Environment variables to set when running "
                              "CMake.",
                              default=DefaultValueValue({})),
    ])# yapf: disable

    # [tool.py-build-cmake.stubgen]
    stubgen = pbc.insert(
        ConfigOption(
            "stubgen",
            "If specified, mypy's stubgen utility will be used to generate "
            "typed stubs for the Python files in the package.",
        ))
    stubgen.insert_multiple([
        ListOfStrConfigOption('packages',
                              "List of packages to generate stubs for, passed "
                              "to stubgen as -p <?>."),
        ListOfStrConfigOption('modules',
                              "List of modules to generate stubs for, passed "
                              "to stubgen as -m <?>."),
        ListOfStrConfigOption('files',
                              "List of files to generate stubs for, passed to "
                              "stubgen without any flags."),
        ListOfStrConfigOption('args',
                              "List of extra arguments passed to stubgen.",
                              default=DefaultValueValue([])),
    ]) # yapf: disable

    # [tool.py-build-cmake.{linux,windows,mac}]
    for system in ["Linux", "Windows", "Mac"]:
        name = system.lower()
        opt = pbc.insert(
            ConfigOption(
                name,
                f"Override options for {system}.",
                create_if_inheritance_target_exists=True,
                default=DefaultValueValue({}),
            ))
        opt.insert_multiple([
            ConfigOption("sdist",
                         f"{system}-specific sdist options.",
                         inherit_from=sdist_pth,
                         create_if_inheritance_target_exists=True),
            ConfigOption("cmake",
                         f"{system}-specific CMake options.",
                         inherit_from=cmake_pth,
                         create_if_inheritance_target_exists=True),
        ])

    # [tool.py-build-cmake.cross]
    cross = pbc.insert(
        ConfigOption(
            "cross",
            "Causes py-build-cmake to cross-compile the project.",
        ))
    cross_pth = get_cross_path()
    cross.insert_multiple([
        StrConfigOption('implementation',
                        "Identifier for the Python implementation.",
                        "implementation = 'cp' # CPython",
                        default=NoDefaultValue('same as current interpreter')),
        StrConfigOption('version',
                        "Python version, major and minor, without dots.",
                        "version = '310' # 3.10",
                        default=NoDefaultValue('same as current interpreter')),
        StrConfigOption('abi',
                        "Python ABI.",
                        "abi = 'cp310'",
                        default=NoDefaultValue('same as current interpreter')),
        StrConfigOption('arch',
                        "Operating system and architecture (no dots or "
                        "dashes, only underscores, all lowercase).",
                        "arch = 'linux_x86_64'",
                        default=NoDefaultValue('same as current interpreter')),
        PathConfigOption('toolchain_file',
                         "CMake toolchain file to use.",
                         default=RequiredValue(),
                         base_path=RelativeToCurrentConfig(project_path),
                         must_exist=not test,
                         allow_abs=True,
                         is_folder=False),
        ListOfStrConfigOption('copy_from_native_build',
                              "If set, this will cause a native version of the "
                              "CMake project to be built and installed in a "
                              "temporary directory first, and the files in this "
                              "list will be copied to the final cross-compiled "
                              "package. This is useful if you need binary "
                              "utilities that run on the build system while "
                              "cross-compiling, or for things like stubs for "
                              "extension modules that cannot be generated while "
                              "cross-compiling.\n"
                              "May include the '*' wildcard "
                              "(but not '**' for recursive patterns)."),
        ConfigOption("sdist",
                     "Override sdist options when cross-compiling.",
                     inherit_from=sdist_pth,
                     create_if_inheritance_target_exists=True),
        ConfigOption("cmake",
                     "Override CMake options when cross-compiling.",
                     inherit_from=cmake_pth,
                     create_if_inheritance_target_exists=True),
    ]) # yapf: disable

    # local override
    root.insert(
        OverrideConfigOption("py-build-cmake.local.toml",
                             "Allows you to override the "
                             "settings in pyproject.toml",
                             targetpath=pbc_pth))

    # cross-compilation local override
    root.insert(
        OverrideConfigOption("py-build-cmake.cross.toml",
                             "Allows you to override the cross-"
                             "compilation settings in pyproject.toml",
                             targetpath=cross_pth))

    return root


from dataclasses import dataclass
from pathlib import Path


@dataclass
class PackageInfo:
    version: str
    package_name: str
    module_name: str


@dataclass
class BuildPaths:
    source_dir: Path
    wheel_dir: Path
    temp_dir: Path
    staging_dir: Path